import platform

from dlt.common import logger
from dlt.destinations.exceptions import (
    DatabaseUndefinedRelation,
)
from dlt.destinations.impl.postgres.sql_client import Psycopg2SqlClient
from dlt.destinations.typing import DBTransaction

from dlt_cratedb.impl.cratedb.utils import SystemColumnWorkaround

if platform.python_implementation() == "PyPy":
    import psycopg2cffi as psycopg2  # type: ignore[import-not-found]
else:
    import psycopg2

from contextlib import contextmanager
from typing import Any, AnyStr, Iterator, List, Optional, Sequence, cast

from dlt.common.destination.dataset import DBApiCursor
from dlt.destinations.sql_client import (
    DBApiCursorImpl,
    raise_database_error,
)


class CrateDbApiCursorImpl(DBApiCursorImpl):
    """
    Compensate for patches by `SystemColumnWorkaround`.
    """

    def _get_columns(self) -> List[str]:
        if self.native_cursor.description:
            return [SystemColumnWorkaround.unquirk(c[0]) for c in self.native_cursor.description]
        return []


class CrateDbSqlClient(Psycopg2SqlClient):
    """
    CrateDB SQL client, mostly compatible with PostgreSQL, with a few deviations.

    - Use `doc` as a default search path.
    - Disable transactions.
    - Apply I/O patches provided by `SystemColumnWorkaround`.
    """

    def open_connection(self) -> "psycopg2.connection":
        """
        Use `doc` instead of `public` as search path with CrateDB.
        """
        self._conn = psycopg2.connect(
            dsn=self.credentials.to_native_representation(),
            options=f"-c search_path={self.fully_qualified_dataset_name()},doc",
        )
        self._reset_connection()
        return self._conn

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        """
        CrateDB does not support transactions. Make emitting `BEGIN TRANSACTION` a no-op.
        """
        logger.warning(
            "CrateDB does not support transactions. Each SQL statement is auto-committed"
            " separately."
        )
        yield self

    @raise_database_error
    def commit_transaction(self) -> None:
        """
        CrateDB does not support transactions. Make emitting `COMMIT` a no-op.
        """
        pass

    @raise_database_error
    def rollback_transaction(self) -> None:
        """
        CrateDB does not support transactions. Raise an exception on `ROLLBACK`.

        TODO: Any better idea?
        """
        raise NotImplementedError("CrateDB statements can not be rolled back")

    # @raise_database_error
    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        """
        Need to patch the result returned from the database.
        """
        result = super().execute_sql(sql, *args, **kwargs)
        return SystemColumnWorkaround.patch_result(result)

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        """
        Need to patch the SQL statement and use the custom `CrateDbApiCursorImpl`.
        """
        query = cast(AnyStr, SystemColumnWorkaround.patch_sql(query))
        curr: DBApiCursor
        db_args = args if args else kwargs if kwargs else None
        with self._conn.cursor() as curr:
            try:
                curr.execute(query, db_args)
                yield CrateDbApiCursorImpl(curr)  # type: ignore[abstract]
            except psycopg2.Error as outer:
                try:
                    self._reset_connection()
                except psycopg2.Error:
                    self.close_connection()
                    self.open_connection()
                raise outer

    @staticmethod
    def _is_error_schema_unknown(exception: Exception) -> bool:
        """
        CrateDB raises `Schema 'testdrive' unknown` errors when accessing schemas without tables.

        psycopg2.errors.InternalError_: Schema 'testdrive_staging' unknown
        dlt.destinations.exceptions.DatabaseTransientException: Schema 'testdrive_staging' unknown

        a) Try to ignore that.
        b) TODO: Refactor to synthesize an empty result, see `job_client_impl.get_stored_state`.
        c) Resolve in CrateDB.
        """
        msg = str(exception)
        return "Schema" in msg and "unknown" in msg

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        """
        Loop in additional error evaluation for CrateDB.
        """
        if cls._is_error_schema_unknown(exception=ex):
            raise DatabaseUndefinedRelation(ex)
        return Psycopg2SqlClient._make_database_exception(ex)

    def create_dataset(self) -> None:
        """
        CrateDB does not know `CREATE|DROP SCHEMA ...` statements.
        A no-op is not enough, because downstream operations expect the schema to exist.
        """
        self.execute_sql(
            f"CREATE TABLE IF NOT EXISTS {self.fully_qualified_dataset_name()}._placeholder (id"
            " INT)"
        )

    def drop_dataset(self) -> None:
        """
        CrateDB does not know `CREATE|DROP SCHEMA ...` statements.
        A no-op is not enough, because downstream operations expect the schema to exist.
        """
        self.execute_sql(f"DROP TABLE IF EXISTS {self.fully_qualified_dataset_name()}._placeholder")



import json
import platform
from datetime import date, datetime, time
from typing import Any, Optional, Sequence, Union

from dlt.common.data_writers.escape import _escape_extended, _make_sql_escape_re

if platform.python_implementation() == "PyPy":
    import psycopg2cffi as psycopg2  # type: ignore[import-not-found]
import psycopg2.sql

# CrateDB does not accept the original `{"'": "''", "\\": "\\\\", "\n": "\\n", "\r": "\\r"}`?
SQL_ESCAPE_DICT = {"'": "''"}
SQL_ESCAPE_RE = _make_sql_escape_re(SQL_ESCAPE_DICT)


def _escape_extended_cratedb(v: str) -> str:
    """
    The best-practice escaper for CrateDB, discovered by trial-and-error.
    """
    return _escape_extended(v, prefix="'", escape_dict=SQL_ESCAPE_DICT, escape_re=SQL_ESCAPE_RE)


def escape_cratedb_literal(v: Any) -> Any:
    """
    Based on `escape_postgres_literal`, with a mix of `escape_redshift_literal`.

    CrateDB needs a slightly adjusted escaping of literals.
    Examples: "L'Aupillon" and "Pizzas d'Anarosa" from `sys.summits`.

    It possibly also doesn't support the `E'` prefix as employed by the PostgreSQL escaper?
    Fortunately, the Redshift escaper came to the rescue, providing a reasonable baseline.

    CrateDB also needs support when serializing container types ARRAY vs. OBJECT.
    """
    if isinstance(v, str):
        return _escape_extended_cratedb(v)
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    # CrateDB, when serializing from an incoming `json` or `jsonb` type, the type mapper
    # can't know about what's actually inside, so arrays need a special treatment.
    if isinstance(v, list):
        v = {"array": v}
    if isinstance(v, dict):
        return _escape_extended_cratedb(json.dumps(v)) + "::OBJECT(DYNAMIC)"
    if isinstance(v, bytes):
        return f"'\\x{v.hex()}'"
    if v is None:
        return "NULL"

    return str(v)


class SystemColumnWorkaround:
    """
    CrateDB reserves `_`-prefixed column names for internal purposes.

    When approaching CrateDB with such private columns which are common in
    ETL frameworks, a corresponding error will be raised.

        InvalidColumnNameException["_dlt_load_id" conflicts with system column pattern]

    This class provides utility methods to work around the problem, brutally.

    See also: https://github.com/crate/crate/issues/15161

    FIXME: Please get rid of this code by resolving the issue in CrateDB.
    """

    quirked_labels = [
        "__dlt_id",
        "__dlt_load_id",
        "__dlt_parent_id",
        "__dlt_list_idx",
        "__dlt_root_id",
    ]

    @staticmethod
    def quirk(thing: str) -> str:
        thing = (
            thing.replace("_dlt_load_id", "__dlt_load_id")
            .replace("_dlt_id", "__dlt_id")
            .replace("_dlt_parent_id", "__dlt_parent_id")
            .replace("_dlt_list_idx", "__dlt_list_idx")
            .replace("_dlt_root_id", "__dlt_root_id")
        )
        return thing

    @staticmethod
    def unquirk(thing: str) -> str:
        thing = (
            thing.replace("__dlt_load_id", "_dlt_load_id")
            .replace("__dlt_id", "_dlt_id")
            .replace("__dlt_parent_id", "_dlt_parent_id")
            .replace("__dlt_list_idx", "_dlt_list_idx")
            .replace("__dlt_root_id", "_dlt_root_id")
        )
        return thing

    @classmethod
    def patch_sql(cls, sql: Union[str, psycopg2.sql.Composed]) -> Union[str, psycopg2.sql.Composed]:
        if isinstance(sql, str):
            sql = cls.quirk(sql)
        elif isinstance(sql, psycopg2.sql.Composed):
            sql.seq[0]._wrapped = cls.quirk(sql.seq[0]._wrapped)
        else:
            raise NotImplementedError(f"Unsupported type '{type(sql)}' for augmenting SQL: {sql}")
        return sql

    @classmethod
    def patch_result(cls, result: Sequence[Sequence[Any]]) -> Optional[Sequence[Sequence[Any]]]:
        if result is None:
            return None
        row_new = []
        for row in result:
            if len(row) >= 2 and row[1] in cls.quirked_labels:
                r = list(row)
                r[1] = cls.unquirk(r[1])
                row = tuple(r)
            row_new.append(row)
        return row_new



import dataclasses
from typing import Final

from dlt.common.configuration import configspec
from dlt.destinations.impl.postgres.configuration import (
    PostgresClientConfiguration,
    PostgresCredentials,
)


@configspec(init=False)
class CrateDbCredentials(PostgresCredentials):
    drivername: Final[str] = dataclasses.field(  # type: ignore[misc]
        default="postgres", init=False, repr=False, compare=False
    )


# CrateDB does not support databases, just schemas.
# In dlt, schemas are conveyed by `dataset_name`?
del CrateDbCredentials.__dataclass_fields__["database"]


@configspec
class CrateDbClientConfiguration(PostgresClientConfiguration):
    destination_type: Final[str] = dataclasses.field(  # type: ignore[misc]
        default="cratedb", init=False, repr=False, compare=False
    )
    credentials: CrateDbCredentials



from dlt_cratedb.patch.utils import info_schema_null_to_bool


def activate_patch() -> None:
    import dlt.destinations.job_client_impl
    import dlt.destinations.utils

    dlt.destinations.job_client_impl.info_schema_null_to_bool = info_schema_null_to_bool
    dlt.destinations.utils.info_schema_null_to_bool = info_schema_null_to_bool



def info_schema_null_to_bool(v: str) -> bool:
    """Converts INFORMATION SCHEMA truth values to Python bool"""
    if v in ("NO", "0", False):
        return False
    elif v in ("YES", "1", True):
        return True
    raise ValueError(v)
